# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

import logging
from dataclasses import field
from typing import List, NewType

from pydantic.dataclasses import dataclass

from conversant.chatbot import Interaction
from conversant.prompts.prompt import Prompt

Conversation = NewType("Conversation", List[Interaction])


@dataclass
class ChatPrompt(Prompt):
    """A chat prompt given to a Chatbot.

    The examples in a `ChatPrompt` are a list of `Conversation`s themselves a list of
    `Interaction`s. This is one level of nesting further as compared to those in
    `Prompt`, which are a list of `Interaction`s.

    Required keys:
        user: An entity speaking to the bot.
        bot: The Chatbot itself.

    Constants:
        REQUIRED_KEYS (List[str]): The list of required keys for the chat prompt.
            (default: `["user", "bot"]`)
        MIN_PREAMBLE_LENGTH (int): The minimum length of the preamble. (default: `1`)
        MIN_NUM_EXAMPLES (int): The minimum number of examples that should be passed in.
            (default: `1`)
    """

    examples: List[Conversation]

    REQUIRED_KEYS: List[str] = field(default_factory=lambda: ["user", "bot"])
    MIN_PREAMBLE_LENGTH: int = 10
    MIN_NUM_EXAMPLES: int = 0

    def __post_init__(self) -> None:
        """Validators for the chat prompt.

        Validates that the prompt follows the requirements of the validators listed
        below. Minimally, the ChatPrompt needs to follow the requirements of its parent
        class.
        """
        super()._validate_preamble()
        super()._validate_example_separator()
        super()._validate_headers()
        self._validate_examples()
        self._validate_dialogue()

    @property
    def user_name(self):
        """
        Returns:
            str: The name of the user that interacts with the chatbot who uses this
                ChatPrompt. Typically this should be set to `'User'`.
        """
        return self.headers["user"]

    @property
    def bot_name(self):
        """
        Returns:
            str: The name of the chatbot who uses this ChatPrompt.
        """
        return self.headers["bot"]

    def create_interaction_string(self, *args, **kwargs) -> str:
        """Creates a string representation of an interaction.

        Interactions will look like the following:

            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n

        Note the colon and space separating the speaker name from the respective
        utterance.

        Args:
            args: Positional arguments for the new interaction.
            kwargs: Keyword arguments for the new interaction.

        Returns:
            str: String representation of an interaction.
        """
        interaction = (
            self.create_interaction(*args, **kwargs) if len(args) > 0 else kwargs
        )
        return "".join(
            f"{self.headers[key]}: {interaction[key]}\n" for key in interaction.keys()
        )

    def create_conversation_string(self, conversation: Conversation) -> str:
        """Creates a string represenation of a conversation.

        Conversations will look like the following:

            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n
            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n

        Args:
            conversation (Conversation): List of interactions.
        """
        return "".join(
            self.create_interaction_string(**interaction)
            for interaction in conversation
        )

    def to_string(self) -> str:
        """Creates a string representation of the conversation prompt.

        The string representation is assembled from the preamble and examples.
        Each example is created from a `create_conversation_string` method and is
        demarcated by an `example_separator`.

        Examples will look like the following:

            {example_separator}
            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n
            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n
            {example_separator}
            {user_name}: {utterance}\n
            {bot_name}: {utterance}\n

        Returns:
            str: String representation of the conversation prompt.
        """
        lines = [f"{self.preamble}\n"]
        lines += self.example_separator + f"{self.example_separator}".join(
            self.create_conversation_string(example) for example in self.examples
        )
        return "".join(lines).strip()

    def _validate_examples(self) -> None:
        """Validates that the `examples` meet the following requirements:

        - All fields are used in every example of `examples`.
        - At least `MIN_NUM_EXAMPLES` examples are given.

        Raises:
            ValueError: If any of the above requirements is not met.
        """
        # All fields are used in every interaction in every example of `examples`.
        for example in self.examples:
            for interaction in example:
                if any(key not in interaction for key in self.REQUIRED_KEYS):
                    raise ValueError(
                        "Missing required key.\nInteraction's keys: "
                        f"{interaction.keys()}\nRequired: {self.REQUIRED_KEYS}"
                    )
        # At least `MIN_NUM_EXAMPLES` examples are given.
        if len(self.examples) < self.MIN_NUM_EXAMPLES:
            raise ValueError(
                f"At least {self.MIN_NUM_EXAMPLES} example(s) must be given for"
                f"{self.__class__.__name__}"
            )

    def _validate_dialogue(self) -> None:
        """Validates that the examples conform to a 2-person dialogue.


        There should only be 2 speakers in the examples, and each speaker's utterance
        should not be prefixed with their name.

        Raises:
            ValueError: If the above requirement is not met.
        """
        for example in self.examples:
            # Only 2 speakers should be in each conversation interaction
            if not all([len(interaction) == 2 for interaction in example]):
                raise ValueError(
                    "Conversation interactions must be pairs of utterances."
                )

            # Only check the examples for name-prefixed utterances if there is at least
            # one interaction
            if example:
                user_turns = [interaction["user"] for interaction in example]
                bot_turns = [interaction["bot"] for interaction in example]
                all_turns = user_turns + bot_turns

                colon_prefixed = all(":" in turn for turn in all_turns)
                hyphen_prefixed = all("-" in turn for turn in all_turns)

                if colon_prefixed or hyphen_prefixed:
                    # This might false-positive, so we only log a warning
                    logging.warning(
                        "Did you mistakenly prefix the example dialogue turns with"
                        "user/bot names?"
                    )

                user_prefixed = all(
                    turn.lstrip().startswith(self.user_name) for turn in user_turns
                )

                bot_prefixed = all(
                    turn.lstrip().startswith(self.bot_name) for turn in bot_turns
                )
                if user_prefixed and bot_prefixed:
                    # It's hard to think of any genuine case where all utterances begin
                    # with self-names.
                    raise ValueError(
                        "Conversation interactions should not be prefixed with user/bot"
                        "names!"
                    )



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

import logging
from typing import Iterable, Optional

import cohere

from conversant.search.document import Document
from conversant.search.searcher import Searcher


class LocalSearcher(Searcher):
    """Searches only a user-provided local document cache."""

    def __init__(self, client: cohere.Client, documents: Iterable[Document]):
        """Loads a searcher with a Cohere client & documents to search over.

        Args:
            client (cohere.Client): Provides access to Cohere API via the Python SDK
            documents (Iterable[Document]): Iterable cache of local Documents.
        """
        super().__init__(client, documents)

    def search(self, query: str, threshold: float = -1e6) -> Optional[Document]:
        """Searches by finding most similar doc in local docs.

        Does not make any updates to local doc cache. Only returns a document
        if the similarity is above the threshold.

        Args:
            query (str): Query to check docs against.
            threshold (float): Minimum similarity needed to return a document.

        Returns:
            Optional[Document]: Most similar doc to query. None if threshold not met.
        """

        embedded_query = self.co.embed(texts=[query]).embeddings[0]
        similarities = [
            self._measure_similarity(embedded_query, d.embedding)
            for d in self.documents
        ]

        max_similarity = max(similarities)

        if max_similarity < threshold:
            logging.warning(
                f"Max search similarity {max_similarity} below threshold {threshold}; "
                "no document returned."
            )
            return None

        logging.info(f"Search result found for query: {query}")
        nearest_idx = similarities.index(max_similarity)
        nearest_doc = self.documents[nearest_idx]

        return nearest_doc



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

import codecs
import pickle
import sys
from typing import Type

import cohere
from streamlit.web import cli as stcli

from conversant.chatbot import Chatbot


def encode_chatbot(chatbot: Type[Chatbot]) -> str:
    """Serialize and encode a Chatbot object to a base-64 string encoding.

    Args:
        chatbot (object): a chatbot of class inherited from Chatbot

    Returns:
        str: Chatbot object as a base-64 string
    """
    chatbot.co = None
    return codecs.encode(pickle.dumps(chatbot), "base64").decode()


def decode_chatbot(chatbot_string: str, client: cohere.Client) -> Type[Chatbot]:
    """Decode and deserialize a Chatbot object.

    Args:
        obj_string (str): a base-64 string encoding

    Returns:
        Type[Chatbot]: a chatbot of class inherited rom Chatbot
    """
    chatbot = pickle.loads(codecs.decode(chatbot_string.encode(), "base64"))
    chatbot.co = client
    return chatbot


def launch_streamlit(chatbot: Type[Chatbot]) -> None:
    """Launches a demo of a chatbot using Streamlit.

    The bot will be a persona available for chatting using the interface
    defined in app/streamlit_example.py.

    Args:
        bot (Type[Chatbot]): a chatbot of class inherited from Chatbot
    """
    sys.argv = "streamlit run app/streamlit_example.py --".split(" ")
    sys.argv.append(encode_chatbot(chatbot))
    sys.exit(stcli.main())



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

import json
import logging
import os
from typing import Any, Dict

import cohere
import jsonschema

from conversant.chatbot import Chatbot, Interaction
from conversant.prompts.chat_prompt import ChatPrompt
from conversant.prompts.prompt import Prompt

PERSONA_MODEL_DIRECTORY = "conversant/personas"
PERSONA_JSON_SCHEMA = {
    "type": "object",
    "properties": {
        "chatbot_config": {
            "type": "object",
            "properties": {
                "max_context_examples": {"type": "integer"},
                "avatar": {"type": "string"},
            },
        },
        "client_config": {
            "type": "object",
            "properties": {
                "model": {"type": "string"},
                "max_tokens": {"type": "integer"},
                "temperature": {"type": "number"},
                "frequency_penalty": {"type": "number"},
                "presence_penalty": {"type": "number"},
                "stop_sequences": {"type": "array"},
            },
        },
        "prompt_config": {
            "type": "object",
        },
    },
}


class PromptChatbot(Chatbot):
    """Use prompt templates and LLM generation to define a chatbot.

    This bot makes no use of external knowledge sources.
    """

    def __init__(
        self,
        client: cohere.Client,
        prompt: Prompt,
        persona_name: str = "",
        chatbot_config: Dict[str, Any] = {},
        client_config: Dict[str, Any] = {},
    ):
        """Enriches init by adding a prompt.

        Args:
            client (cohere.Client): Cohere client for API
            prompt (Prompt): Prompt object to direct behavior.
            persona_name (str, optional): Bot's persona name. Defaults to empty string.
            chatbot_config: (Dict[str, Any], optional): Bot's chat config. Defaults to
                empty dict.
            client_config (Dict[str, Any], optional): Bot's client config. Defaults to
                empty dict.
        """

        super().__init__(client)
        self.prompt = prompt
        self.persona_name = persona_name

        self.configure_chatbot(chatbot_config)
        self.configure_client(client_config)
        self.chat_history = []
        self.prompt_history = [self.prompt.to_string()]

    def __repr__(self) -> str:
        return json.dumps(self.to_dict(), indent=4, default=str)

    @property
    def user_name(self):
        """
        Returns:
            str: The name of the user, defined in the prompt. Defaults to "User".
        """
        if hasattr(self.prompt, "user_name"):
            return self.prompt.user_name
        else:
            return "User"

    @property
    def bot_name(self):
        """
        Returns:
            str: The name of the chatbot, defined in the prompt. Defaults to
                "PromptChatbot".
        """
        if hasattr(self.prompt, "bot_name"):
            return self.prompt.bot_name
        else:
            return "PromptChatbot"

    @property
    def latest_prompt(self) -> str:
        """Retrieves the latest prompt.

        Returns:
            str: The prompt most recently added to the prompt history.
        """
        return self.prompt_history[-1]

    def reply(self, query: str) -> Interaction:
        """Replies to a query given a chat history.

        The reply is then generated directly from a call to a LLM.

        Args:
            query (str): A query passed to the prompt chatbot.

        Returns:
            Interaction: Dictionary of query and generated LLM response
        """
        # The current prompt is assembled from the initial prompt,
        # from the chat history with a maximum of max_context_examples,
        # and from the current query
        current_prompt = self.get_current_prompt(query)

        # Make a call to Cohere's co.generate API
        generated_object = self.co.generate(
            model=self.client_config["model"],
            prompt=current_prompt,
            max_tokens=self.client_config["max_tokens"],
            temperature=self.client_config["temperature"],
            frequency_penalty=self.client_config["frequency_penalty"],
            presence_penalty=self.client_config["presence_penalty"],
            stop_sequences=self.client_config["stop_sequences"],
        )

        # If response was cut off by .generate() finding a stop sequence,
        # remove that sequence from the response.
        response = generated_object.generations[0].text
        for stop_seq in self.client_config["stop_sequences"]:
            if response.endswith(stop_seq):
                response = response[: -len(stop_seq)]
        response = response.lstrip()

        # We need to remember the current response in the chat history for future
        # responses.
        self.chat_history.append(self.prompt.create_interaction(query, response))
        self.prompt_history.append(current_prompt)

        return response

    def get_current_prompt(self, query) -> str:
        """Stitches the prompt with a trailing window of the chat.

        Args:
            query (str): The current user query.

        Returns:
            str: The current prompt given a query.
        """
        # get base prompt
        base_prompt = self.prompt.to_string() + "\n"

        # get context prompt
        context_prompt_lines = []
        trimmed_chat_history = (
            self.chat_history[-self.chatbot_config["max_context_examples"] :]
            if self.chatbot_config["max_context_examples"] > 0
            else []
        )
        # TODO when prompt is updated, the history is mutated
        # as it is recreated using the new prompt. A possible fix is to save the old
        # prompt in history and use it when recreating.
        for turn in trimmed_chat_history:
            context_prompt_lines.append(self.prompt.create_interaction_string(**turn))
        context_prompt = self.prompt.example_separator + "".join(context_prompt_lines)

        # get query prompt
        query_prompt = self.prompt.create_interaction_string(query)

        current_prompt = base_prompt + context_prompt + query_prompt
        return current_prompt.strip()

    def configure_chatbot(self, chatbot_config: Dict = {}) -> None:
        """Configures chatbot options.

        Args:
            chatbot_config (Dict, optional): Updates self.chatbot_config. Defaults
                to {}.
        """
        # We initialize the chatbot to these default config values.
        if not hasattr(self, "chatbot_config"):
            self.chatbot_config = {"max_context_examples": 10, "avatar": ":robot:"}
        # Override default config values with the config passed in
        if isinstance(chatbot_config, Dict):
            self.chatbot_config.update(chatbot_config)
        else:
            raise TypeError(
                "chatbot_config must be of type Dict, but was passed in as "
                f"{type(chatbot_config)}"
            )

    def configure_client(self, client_config: Dict = {}) -> None:
        """Configures client options.

        Args:
            client_config (Dict, optional): Updates self.client_config. Defaults to {}.
        """
        # We initialize the client to these default config values.
        if not hasattr(self, "client_config"):
            self.client_config = {
                "model": "xlarge",
                "max_tokens": 100,
                "temperature": 0.75,
                "frequency_penalty": 0.0,
                "presence_penalty": 0.0,
                "stop_sequences": ["\n"],
            }
        # Override default config values with the config passed in
        if isinstance(client_config, Dict):
            self.client_config.update(client_config)
        else:
            raise TypeError(
                "client_config must be of type Dict, but was passed in as "
                f"{type(client_config)}"
            )

    @classmethod
    def from_persona(
        cls,
        persona_name: str,
        client: cohere.Client,
        persona_dir: str = PERSONA_MODEL_DIRECTORY,
    ):
        """Initializes a PromptChatbot using a persona.

        Args:
            persona (str): Name of persona, corresponding to a .json file.
            client (cohere.Client): Cohere client for API
            persona_dir (str): Path to where pre-defined personas are.
        """
        # Load the persona from a local directory
        persona_path = os.path.join(persona_dir, persona_name, "config.json")
        if os.path.isfile(persona_path):
            logging.info(f"loading persona from {persona_path}")
        else:
            raise FileNotFoundError(f"{persona_path} cannot be found.")
        with open(persona_path) as f:
            persona = json.load(f)

        # Validate that the persona follows our predefined schema
        cls._validate_persona_dict(persona, persona_path)

        return cls(
            client=client,
            prompt=ChatPrompt.from_dict(persona["chat_prompt_config"]),
            persona_name=persona_name,
            chatbot_config=persona["chatbot_config"],
            client_config=persona["client_config"],
        )

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this instance into a Python dictionary.

        Returns:
            Dict[str, Any]: Dictionary of attributes that defines this instance of a
                PromptChatbot.
        """
        return {
            "co": self.co,
            "prompt": self.prompt.to_dict(),
            "persona_name": self.persona_name,
            "chatbot_config": self.chatbot_config,
            "client_config": self.client_config,
            "chat_history": self.chat_history,
            "prompt_history": self.prompt_history,
            "user_name": self.user_name,
            "bot_name": self.bot_name,
            "latest_prompt": self.latest_prompt,
        }

    @staticmethod
    def _validate_persona_dict(persona: Dict[str, Any], persona_path: str) -> None:
        """Validates formatting of a persona defined as a dictionary.

        Args:
            persona (Dict[str, Any]): A dictionary containing the persona.
            persona_path: The path from which the persona was loaded.
        """
        try:
            jsonschema.validate(instance=persona, schema=PERSONA_JSON_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise jsonschema.exceptions.ValidationError(
                f"Type of values in given dictionary (persona from {persona_path}) do "
                f"not match schema': {e}"
            )
        except KeyError as e:
            raise KeyError(
                f"Invalid key in given dictionary (persona from {persona_path})': {e}"
            )
        except Exception as e:
            raise Exception(
                "Failed to validate persona in given dictionary (persona from "
                f"{persona_path}): {e}"
            )



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

import json
from dataclasses import field
from typing import Any, Dict, List

from pydantic.dataclasses import dataclass

from conversant.chatbot import Interaction


@dataclass
class Prompt:
    """Base class for all structured zero-shot or few-shot prompts.

    Args:
        preamble (str): A string that directs the model to behave in certain ways by
            describing its function (e.g. a description of a bot's persona).
        example_separator (str): A separator for each example.
        headers (Dict[str, str]): A dictionary mapping from keys in examples to the
            values that will substitute them.
        examples (List[Interaction]): A list of examples to illustrate the intended
            behaviour.

    Constants:
        REQUIRED_HEADER_KEYS (List[str]): The list of required keys in headers for the
            prompt. (default: `[]`)
        MIN_PREAMBLE_LENGTH (int): The minimum length of the preamble. (default: `1`)
        MIN_NUM_EXAMPLES (int): The minimum number of examples that should be passed in.
            (default: `1`)
    """

    preamble: str
    example_separator: str
    headers: Dict[str, str]
    examples: List[Interaction]

    REQUIRED_KEYS: List[str] = field(default_factory=lambda: [])
    MIN_PREAMBLE_LENGTH: int = 1
    MIN_NUM_EXAMPLES: int = 1

    def __post_init__(self) -> None:
        """Validators for each prompt.

        Each subclass that inherits from Prompt should call this using
        `super().__post_init__()` so that their prompt structure is also validated.
        Stricter validation can be implemented in subclasses by overriding these
        methods, defining custom validators, or adjusting the constants of Prompt.
        """
        self._validate_preamble()
        self._validate_example_separator()
        self._validate_headers()
        self._validate_examples()

    def __repr__(self) -> str:
        return self.to_string()

    def __str__(self) -> str:
        return self.to_string()

    def create_interaction(self, *args, **kwargs) -> Interaction:
        """Creates a new dictionary representation of an interaction.

        The order of args here should correspond to the order of the keys in `headers`.
        The i-th positional argument passed in corresponds to the i-th key, up to
        `len(headers)`. If fewer than `len(headers)` arguments are passed in, the
        remaining entries default to `""`. If more than `len(headers)` arguments are
        passed in, they are ignored.

        Any subsequent keyword arguments override the values defined by the positional
        arguments.

        Args:
            args: Positional arguments for the new interaction.
            kwargs: Keyword arguments for the new interaction.

        Returns:
            Interaction: Dictionary representation of an interaction.
        """
        new_interaction = {
            key: args[i] if i < len(args) else ""
            for i, key in enumerate(self.headers.keys())
        }
        new_interaction.update(kwargs)
        return new_interaction

    def create_interaction_string(self, *args, **kwargs) -> str:
        """Creates a string representation of an interaction.

        The order of args here should correspond to the order of the keys in `headers`.
        The i-th positional argument passed in corresponds to the i-th key, up to
        `len(headers)`. If fewer than `len(headers)` arguments are passed in, the
        remaining entries default to `""`. If more than `len(headers)` arguments are
        passed in, they are ignored.

        Any subsequent keyword arguments override the values defined by the positional
        arguments.

        Each prompt can have their own way of stitching together headers and values
        within examples. Generally, each header should follow its corresponding example
        value. If there are no positional arguments passed in, then the ordering of the
        variables in examples follows the order of the keyword arguments. Otherwise, a
        new example dictionary is created from the positional arguments and the ordering
        is dependent on the order of the `headers`.

        Interactions will look like the following:

            {header}{value}\n
            {header}{value}\n

        Any custom logic should be defined in a subclass method that
        overrides this method.

        Args:
            args: Positional arguments for the new interaction.
            kwargs: Keyword arguments for the new interaction.

        Returns:
            str: String representation of an interaction.
        """
        interaction = self.create_interaction(*args, **kwargs) if args else kwargs
        return "".join(
            f"{self.headers[key]}{interaction[key]}\n" for key in interaction.keys()
        )

    def to_string(self) -> str:
        """Creates a string representation of the prompt.

        The string representation is assembled from the preamble and examples.
        Each example is created from a `create_interaction_string` method and is
        demarcated by an `example_separator`.

        Examples will look like the following:

            {preamble}\n
            {example_separator}
            {field}{value}\n
            {field}{value}\n
            {example_separator}
            {field}{value}\n
            {field}{value}\n
            ...

        Returns:
            str: String representation of the prompt.
        """
        lines = [f"{self.preamble}\n"]
        lines += self.example_separator + f"{self.example_separator}".join(
            self.create_interaction_string(**example) for example in self.examples
        )
        return "".join(lines).strip()

    def update(self, config: Dict[str, Any]) -> None:
        """Updates attributes of this class with attributes from `config`.

        Args:
            config (Dict[str, Any]): Dictionary of attributes that should be updated for
                this class.
        """
        for key, value in config.items():
            if hasattr(self, key):
                setattr(self, key, value)

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "Prompt":
        """Instantiates a `Prompt` from a dictionary.

        Args:
            config (Dict[str, Any]: Dictionary used to instantiate a prompt object.
            The dictionary should have the following required keys: `preamble`,
            `headers`, `examples`, example_separator`

        Returns:
            Prompt: The prompt object instantiated from the `config`.
        """
        return cls(**config)

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this instance into a Python dictionary.

        Returns:
            Dict[str, Any]: Dictionary of attributes that defines this instance of a
                Prompt.
        """
        return {
            "preamble": self.preamble,
            "example_separator": self.example_separator,
            "headers": self.headers,
            "examples": self.examples,
        }

    def to_json_string(self) -> str:
        """Serializes this instance into a JSON string.

        Returns:
            str: JSON string representation of this instance of a Prompt.
        """
        prompt_dict = self.to_dict()
        return json.dumps(prompt_dict, indent=4) + "\n"

    def _validate_preamble(self) -> None:
        """Validates that the preamble meets the following requirements:

        - At least `MIN_PREAMBLE_LENGTH` in length.

        Raises:
            ValueError: If the length of the preamble is less than
                `MIN_PREAMBLE_LENGTH`.
        """
        if len(self.preamble) < self.MIN_PREAMBLE_LENGTH:
            raise ValueError(
                f"Preamble must be at least {self.MIN_PREAMBLE_LENGTH} characters."
            )

    def _validate_headers(self) -> None:
        """Validates that `headers` meets the following requirements:

        - Contains all keys in `REQUIRED_KEYS`.

        Raises:
            ValueError: If any keys in `REQUIRED_KEYS` is missing from the prompt's
                `headers`.
        """
        if any(key not in self.headers.keys() for key in self.REQUIRED_KEYS):
            raise ValueError(
                f"Missing required key.\nHeader keys: {self.headers.keys()}.\n"
                f"Required: {self.REQUIRED_KEYS}."
            )

    def _validate_example_separator(self) -> None:
        """Validates that the `example_separator` meets the following requirements:

        - Is a str.

        Raises:
            TypeError: If the `example_separator` is not a `str`.
        """
        if not isinstance(self.example_separator, str):
            raise ValueError(
                "example_separator must be a string. Current type: "
                f"{type(self.example_separator)}"
            )

    def _validate_examples(self) -> None:
        """Validates that the `examples` meet the following requirements:

        - All keys are used in every example of `examples`.
        - At least `MIN_NUM_EXAMPLES` examples are given.

        Raises:
            ValueError: If any of the above requirements is not met.
        """
        # All required keys are used in every example of `examples`.
        for example in self.examples:
            if any(key not in example for key in self.REQUIRED_KEYS):
                raise ValueError(
                    f"Missing required key.\nHeader keys: {self.headers.keys()}.\n"
                    f"Required: {self.REQUIRED_KEYS}."
                )

        # At least `MIN_NUM_EXAMPLES` examples are given.
        if len(self.examples) < self.MIN_NUM_EXAMPLES:
            raise ValueError(
                f"At least {self.MIN_NUM_EXAMPLES} example must be given for "
                f"{self.__class__.__name__}"
            )



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Document:
    """Schema for documents retrieved by searchers."""

    source_link: str
    doc_id: str
    content: str
    embedding: Optional[List[float]] = None



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

from abc import ABC, abstractmethod
from typing import Dict, List, NewType

import cohere

Interaction = NewType("Interaction", Dict[str, str])


class Chatbot(ABC):
    """Defines bots that reply to users in a 1:1 text conversation"""

    def __init__(self, client: cohere.Client):
        """A Chatbot should be passed a Cohere Client for API calls.

        Args:
            client (cohere.Client): Provides access to Cohere API via the Python SDK
        """

        self.co = client

        # Holds the full, formatted chat history with type of
        # List[Interaction]. Each dict object represents a
        # conversation turn and has two keys: "speaker_name" and
        # "utterance".
        self.chat_history: List[Interaction] = []

    @abstractmethod
    def reply(self, query: str) -> Interaction:
        """Replies to a user given some input and context.

        Args:
            query (str): Most recent message from the user.

        Returns:
            Interaction: A dictionary consisting of the interaction, including the
                query and response.
        """



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

from abc import ABC, abstractmethod
from typing import Iterable, List, Optional

import cohere
import numpy as np

from conversant.search.document import Document


class Searcher(ABC):
    """Searches through documents for ground truth."""

    def __init__(self, client: cohere.Client, documents: Iterable[Document]):
        """Loads a searcher with a Cohere client & documents to search over.

        Args:
            client (cohere.Client): Provides access to Cohere API via the Python SDK
            documents (Iterable[Document]): Iterable cache of local Documents.
        """
        self.co = client
        self.documents = documents

    @abstractmethod
    def search(self, query: str) -> Optional[Document]:
        """Semantic search over local documents.

        Precisely _how_ the cache of local documents is updated
        is left to the implementation in a subclass.
        Args:
            query (str): Input query to match documents against.
        Returns:
            Optional[Document]: Best matching document, if found.
        """

    def _measure_similarity(self, embed_1: List[float], embed_2: List[float]) -> float:
        """Measures similarity between embeddings. Uses cosine similarity.

        Manually computes dot(A,B)/(||A||*||B||), to avoid adding
        more deps. Uses numpy methods for these computations.

        Args:
            embed_1 (List[float]): First embedding in the pair.
            embed_2 (List[float]): Second embedding in the pair.
        Returns:
            float: Similarity between the documents.
        """
        dot_product = np.dot(embed_1, embed_2)
        norm_product = np.linalg.norm(embed_1) * np.linalg.norm(embed_2)

        cos_similarity = dot_product / norm_product

        return cos_similarity

    def embed_documents(self) -> None:
        """Embeds the Searcher's documents."""

        embeddings = self.co.embed(texts=[d.content for d in self.documents]).embeddings

        for document, embedding in zip(self.documents, embeddings):
            document.embedding = embedding



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.


from dataclasses import field
from typing import List

from pydantic.dataclasses import dataclass

from conversant.prompts.prompt import Prompt


@dataclass
class RewritePrompt(Prompt):
    """A rewrite prompt given to a Chatbot.

    Required keys:
        conversation: The possibly ungrounded message to be rewritten.
        fact: A reference paragraph containing a fact to ground the message.
        rewrite: A rewritten grounded message based on the reference fact.

    Constants;
        REQUIRED_KEYS (List[str]): The list of required keys for the prompt. (default:
            `["conversation", "rewrite", "fact"]`)
        MIN_PREAMBLE_LENGTH (int): The minimum length of the preamble. (default: `10`)
        MIN_NUM_EXAMPLES (int): The minimum number of examples that should be passed in.
            (default: `1`)
    """

    REQUIRED_KEYS: List[str] = field(
        default_factory=lambda: ["conversation", "rewrite", "fact"]
    )
    MIN_PREAMBLE_LENGTH: int = 10
    MIN_NUM_EXAMPLES: int = 1

    def __post_init__(self) -> None:
        """Validators for the rewrite prompt.

        Validates that the prompt follows the requirements of the validators listed
        below. Minimally, the RewritePrompt needs to follow the requirements of its
        parent class.
        """
        super().__post_init__()

    def create_interaction_string(self, *args, **kwargs) -> str:
        """Creates a string representation of a grounded rewriting interaction.

        Interactions will look like the following:

            {conversation_header}\n
            {conversation}\n
            {fact_header}\n
            {fact}\n
            {rewrite_header}\n
            {rewrite}\n

        Args:
            args: Positional arguments for the new interaction.
            kwargs: Keyword arguments for the new interaction.

        Returns:
            str: String representation of an interaction.
        """
        interaction = self.create_interaction(*args, **kwargs) if args else kwargs
        return "".join(
            f"{self.headers[key]}\n{interaction[key]}\n" for key in interaction.keys()
        )



# Copyright (c) 2022 Cohere Inc. and its affiliates.
#
# Licensed under the MIT License (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License in the LICENSE file at the top
# level of this repository.

from conversant.search.document import Document
from conversant.search.local_searcher import LocalSearcher
from conversant.search.searcher import Searcher

__all__ = ["Document", "Searcher", "LocalSearcher"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['conversant', 'conversant.prompts', 'conversant.search', 'conversant.utils']

package_data = \
{'': ['*'],
 'conversant': ['personas/client-support/*',
                'personas/fantasy-wizard/*',
                'personas/fortune-teller/*',
                'personas/injured-person/*',
                'personas/math-teacher/*',
                'personas/personal-trainer/*',
                'personas/watch-sales-agent/*']}

install_requires = \
['cohere>=2.8.0,<3.0.0',
 'emoji>=2.1.0,<3.0.0',
 'emojificate>=0.6.0,<0.7.0',
 'pydantic>=1.10.2,<2.0.0',
 'streamlit-ace>=0.1.1,<0.2.0',
 'streamlit-talk>=v0.0.3,<0.0.4',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'conversant',
    'version': '0.1.1',
    'description': 'Conversational AI tooling',
    'long_description': '```\n################################################################################\n#    ____      _                     ____                  _ _                 #\n#   / ___|___ | |__   ___ _ __ ___  / ___|  __ _ _ __   __| | |__   _____  __  #\n#  | |   / _ \\| \'_ \\ / _ \\ \'__/ _ \\ \\___ \\ / _` | \'_ \\ / _` | \'_ \\ / _ \\ \\/ /  #\n#  | |__| (_) | | | |  __/ | |  __/  ___) | (_| | | | | (_| | |_) | (_) >  <   #\n#   \\____\\___/|_| |_|\\___|_|  \\___| |____/ \\__,_|_| |_|\\__,_|_.__/ \\___/_/\\_\\  #\n#                                                                              #\n# This project is part of Cohere Sandbox, Cohere\'s Experimental Open Source    #\n# offering. This project provides a library, tooling, or demo making use of    #\n# the Cohere Platform. You should expect (self-)documented, high quality code  #\n# but be warned that this is EXPERIMENTAL. Therefore, also expect rough edges, #\n# non-backwards compatible changes, or potential changes in functionality as   #\n# the library, tool, or demo evolves. Please consider referencing a specific   #\n# git commit or version if depending upon the project in any mission-critical  #\n# code as part of your own projects.                                           #\n#                                                                              #\n# Please don\'t hesitate to raise issues or submit pull requests, and thanks    #\n# for checking out this project!                                               #\n#                                                                              #\n################################################################################\n```\n\n**Maintainer:** [Cohere ConvAI Team](mailto:convai@cohere.com) \\\n**Project maintained until at least (YYYY-MM-DD):** 2023-03-01\n\n# Conversant\n[![tests](https://github.com/cohere-ai/sandbox-conversant-lib/actions/workflows/run_tests.yaml/badge.svg)](https://github.com/cohere-ai/sandbox-conversant-lib/actions/workflows/run_tests.yaml/badge.svg)\n[![PyPI](https://img.shields.io/pypi/v/conversant.svg)](https://img.shields.io/pypi/v/conversant.svg)\n[![License: MIT](https://img.shields.io/badge/License-MIT-green.svg)](https://opensource.org/licenses/MIT)\n\nBuild conversational AI on top of [Cohere](https://cohere.ai/)\'s [large language models](https://docs.cohere.ai/generate-reference/)\n- 🗣 Use large language models quickly with Cohere\'s [API](https://docs.cohere.ai/api-reference/)\n- 😃 Customize personas\n- 💁 Leave management of chat history up to `conversant`\n- 🪨 Ground conversations in your provided facts (soon!)\n- 🐍 Install `conversant` with pip\n\n`conversant` is a work-in-progress framework for building customizable dialogue agents (aka chatbots) that can answer questions and converse with users with a variety of different chatbot personas. `conversant` aims\nto be modular, flexible and extensible so you can create any kind of chatbots you want!\n\nWe provide several custom personas for you, including 🧑\u200d💼 a client support agent, ⌚️ a watch sales agent, 🧑\u200d🏫 a math teacher, and 🧙 a fantasy wizard. Create your own persona with just a description and some example conversations!\n\n<!-- Read more about `conversant` on our [launch blog post]() and [threads](), and how it is part of the [Cohere Sandbox](). -->\n\nTry `conversant` on our Streamlit demo [here](https://cohere-ai-sandbox-conversant-lib-appstreamlit-example-hgyk4x.streamlitapp.com/)! 🎉\n\n## Table of Contents\n1. [Installation and Usage](#installation-and-usage)\n   1. [Installation](#installation)\n   2. [Streamlit Demo](#streamlit-demo)\n   3. [Running Your Own Demo Locally](#running-your-own-demo-locally)\n   4. [Creating a Custom Persona](#creating-a-custom-persona)\n   5. [Editing a Persona on the Demo](#editing-a-persona-on-the-demo)\n   6. [Usage](#usage)\n2. [How Conversant Works](#how-conversant-works)\n3. [Documentation](#documentation)\n4. [Get Support](#get-support)\n5. [Contributing Guidelines](#contributing-guidelines)\n6. [License](#license)\n\n## Installation and Usage\n\n### Installation\n\n`conversant` is available [on PyPI](https://pypi.org/project/conversant/), and is tested on Python 3.8+ and [Cohere](https://pypi.org/project/cohere/) 2.8.0+.\n```\npip install conversant\n```\n### Streamlit Demo\n\nWant to see it in action first? You can use `conversant` on a [Streamlit](https://docs.streamlit.io/) app without installing anything [here](https://cohere-ai-sandbox-conversant-lib-appstreamlit-example-hgyk4x.streamlitapp.com/)! 🎉\n\n<p float="none">\n  <img src="static/fortune-teller-setup.png" alt="Screenshot showing the available personas on the Streamlit demo, with the Fortune Teller persona selected by default.." height="550"/>\n  <img src="static/fortune-teller-chat.png" alt="Screenshot showing an exchange between a Fortune Teller chatbot and a user." height="550"/>\n</p>\n\n### Running Your Own Demo Locally\n\nCohere uses Streamlit to create its demo applications. If you’re new to Streamlit, you can install it [here](https://docs.streamlit.io/library/get-started/installation) and read more about running Streamlit commands [here](https://docs.streamlit.io/library/get-started/main-concepts).\n\nIf you\'d like to spin up your own instance of the Streamlit demo, you will first need a `COHERE_API_KEY`. \nYou can generate one by visiting [dashboard.cohere.ai](https://dashboard.cohere.ai). \nAdd the key to `.streamlit/secrets.toml`:\n```\nCOHERE_API_KEY = "YOUR_API_KEY_HERE"\n```\n\n Start the Streamlit app from `app/streamlit_example.py`:\n```\nstreamlit run app/streamlit_example.py\n```\n\n### Creating a Custom Persona\nOnce you have your own instance of the Streamlit app, you can begin experimenting with creating custom personas! Check out the `config.json` for each persona in [`conversant/personas`](/conversant/personas/) directory. You\'ll need to create a subfolder within this directory that corresponds to your new persona and add a `config.json` file. The directory structure should look like this:\n\n```\nconversant/personas\n├── fortune-teller\n│   └── config.json\n└── your-persona-name # new\n    └── config.json\n```\n\nThe config file should contain the following:\n- `chatbot_config`: \n  -  `max_context_examples`: The length of the chat history for the chatbot to use in reply.\n  -  `avatar`: Optional emoji shortcode or URL to image as the chatbot\'s avatar. Defaults to 🤖.\n-  `client_config`: Parameters for [`co.generate()`](https://docs.cohere.ai/generate-reference)\n-  `chat_prompt_config`: \n   - `preamble`: Description of the persona.\n   - `example_separator`: A string that separates each example conversation.\n   - `headers`: A name for the `bot` and the `user`.\n   - `examples`: A few conversation examples (few-shot), or empty (zero-shot).\n\n`conversant` will take care of the rest! As an example, check out [`fortune-teller/config.json`](/conversant/personas/fortune-teller/config.json). When you launch the Streamlit app, the new persona will appear in the drop down menu.\n\n### Editing a Persona on the Demo\nYou can also edit a persona on the Streamlit app!\n<img src="static/fortune-teller-edit.png" alt="Screenshot showing the interface for editing a persona on the Streamlit app."/>\n\n### Usage\n\nWith `conversant`, you can create a chatbot powered by [Cohere](https://cohere.ai/)\'s large language models with just the following code snippet.\n```python\nimport cohere\n\nco = cohere.Client("YOUR_API_KEY_HERE")\nbot = conversant.PromptChatbot.from_persona("fantasy-wizard", client=co)\nprint(bot.reply("Hello!"))\n>>> "Well met, fair traveller. What bringest thou to mine village?"\n```\n\nYou can also define your own persona by passing in your own `ChatPrompt`. \n```python\nfrom conversant.prompts import ChatPrompt\n\nshakespeare_config = {\n    "preamble": "Below is a conversation between Shakespeare and a Literature Student.",\n    "example_separator": "<CONVERSATION>\\n",\n    "headers": {\n        "user": "Literature Student",\n        "bot": "William Shakespeare",\n    },\n    "examples": [\n        [\n            {\n                "user": "Who are you?",\n                "bot": "Mine own nameth is Shakespeare, and I speaketh in riddles.",\n            },\n        ]\n    ],\n}\nshakespeare_bot = PromptChatbot(\n    client=co, prompt=ChatPrompt.from_dict(shakespeare_config)\n)\nprint(shakespeare_bot.reply("Hello!"))\n>>> "Greeteth, and welcome. I am Shakespeare, the great poet, dramatist, and playwright."\n```\n\nFrom here, it\'s also possible to talk to your chatbot using the [Streamlit](https://docs.streamlit.io/) app! This will launch the demo with your chatbot persona pre-selected.\n```python\nfrom conversant.utils import demo_utils\ndemo_utils.launch_streamlit(shakespeare_bot)\n```\n## How Conversant Works\n`conversant` uses prompt completion to define a chatbot persona with a description and a few examples. The prompt is sent as input to Cohere\'s [`co.generate()`](https://docs.cohere.ai/generate-reference/) endpoint for an autoregressive language model to generate text in a few-shot manner from the examples and the current dialogue context. \n\nEach user message and chatbot response is appended to a chat history so that future responses are conditioned on the dialogue context at that point in time.\n\nIn the future, we plan to add functionality for a chatbot to be factually grounded using text that is retrieved from a local document cache.\n\nFor more information, refer to [this section in `CONTRIBUTORS.md`](CONTRIBUTORS.md#conversant-schematic).\n\n## Documentation\nFull documentation can be found [here](https://cohere-ai.github.io/sandbox-conversant-lib/).\n\n## Get Support\n\nIf you have any questions or comments, please file an issue or reach out to us on [Discord](discord.gg/co-mmunity).\n\n## Contributing Guidelines\nIf you would like to contribute to this project, please read [`CONTRIBUTORS.md`](CONTRIBUTORS.md)\nin this repository, and sign the Contributor License Agreement before submitting\nany pull requests. A link to sign the Cohere CLA will be generated the first time \nyou make a pull request to a Cohere repository.\n\nIn addition to guidelines around submitting code to this repository, [`CONTRIBUTORS.md`](CONTRIBUTORS.md) contains a walkthrough to help developers get started, as well as schematics that explain how `conversant` works under the hood. :wrench:\n\n## License\n`conversant` has an [MIT License](LICENSE).\n',
    'author': 'Cohere ConvAI',
    'author_email': 'convai@cohere.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/cohere-ai/sandbox-conversant-lib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
